import * as express from 'express';
import express__default, { Request } from 'express';
import nodemailer from 'nodemailer';
import winston from 'winston';
import { BetterSQLite3Database } from 'drizzle-orm/better-sqlite3';
import * as node_mocks_http from 'node-mocks-http';
import JSONTransport from 'nodemailer/lib/json-transport';
import Mail from 'nodemailer/lib/mailer';
import * as drizzle_orm_sqlite_core from 'drizzle-orm/sqlite-core';
import { BaseSQLiteDatabase } from 'drizzle-orm/sqlite-core';
import BetterSqlite3Database from 'better-sqlite3';
import { ExtractTablesWithRelations } from 'drizzle-orm';

type MiddlewareStackArgs = {
    app: express__default.Express;
    config: ExpandedPlainwebConfig<Record<string, unknown>, MailConfig | undefined>;
};
type HttpConfig<T> = {
    port?: number;
    redirects?: Record<string, string>;
    staticPath: string;
    middleware: (opts: MiddlewareStackArgs) => Promise<void> | void;
};
type SmtpMailConfig = {
    host: string;
    port: number;
    secure: boolean;
    auth: {
        user: string;
        pass: string;
    };
};
type MailerConfig = SmtpMailConfig | nodemailer.Transporter;
type MailConfig = {
    default: MailerConfig;
    [key: string]: MailerConfig;
};
type DatabaseConfig<T extends Record<string, unknown>> = {
    dbUrl: string;
    pragma: Record<string, string>;
    schema: T;
    migrationsTable?: string;
};
type PathsConfig = {
    routes?: string;
    cli?: string;
    tasks?: string;
    migrations?: string;
    forms?: string;
    public?: string;
    schema?: string;
};
type LoggerConfig = {
    level?: "silly" | "debug" | "verbose" | "http" | "info" | "warn" | "error";
    transports?: winston.transport[];
    logger?: winston.Logger;
};
type PlainWebConfig<S extends Record<string, unknown>, M extends MailConfig | undefined = undefined> = {
    nodeEnv: "development" | "production" | "test";
    http: HttpConfig<S>;
    database: DatabaseConfig<S>;
    logger?: LoggerConfig;
    paths?: PathsConfig;
    mail?: M;
};
type ExpandedPlainwebConfig<S extends Record<string, unknown>, M extends MailConfig | undefined = undefined> = {
    nodeEnv: "development" | "production" | "test";
    http: Required<HttpConfig<S>>;
    database: Required<DatabaseConfig<S>>;
    logger: Omit<Required<LoggerConfig>, "logger"> & {
        logger?: winston.Logger;
    };
    paths: Required<PathsConfig>;
    mail: M extends undefined ? Record<string, never> : M;
};
/** Define a plainweb configuration */
declare function defineConfig<T extends Record<string, unknown>, M extends MailConfig | undefined = undefined>(config: PlainWebConfig<T, M>): ExpandedPlainwebConfig<T, M>;

/** Return a logger instances with an optional name. Return a prod logger if NODE_ENV=production, otherwise return a dev logger. */
declare function getLogger(name?: string): winston.Logger;
/** A pre-configured logger instance with the name "app". */
declare const log: winston.Logger;

/** Return a random cuid2 id with an optional prefix.
 * @example
 * randomId("usr") // usr_1a2b3c4d
 * randomId() // 1a2b3c4d
 */
declare function randomId(prefix?: string): string;

interface FlashMessage {
    type: "success" | "error";
    message: string;
}

interface PlainResponse {
    _tag: "PlainResponse";
    headers?: Record<string, string>;
    status?: number;
    body?: string | Promise<string>;
    htmlStream?: (rid: number | string) => JSX.Element;
}
/** Return a HTML response with Content-Type set to text/html. */
declare function html(html: Promise<string> | string, opts?: {
    status?: number;
}): PlainResponse;
/**
 * Stream a HTML response with Content-Type set to text/html.
 * Cache-Control is set to no-transform to prevent caching.
 */
declare function stream(htmlStream: (rid: number | string) => JSX.Element): PlainResponse;
/**
 * Redirect to a new path.
 * The status code defaults to 302.
 */
declare function redirect$1(path: string, opts?: {
    message?: FlashMessage;
    status?: number;
}): PlainResponse;
/** Return a JSON response with Content-Type set to application/json. */
declare function json$1(json: JSONSerializable, opts?: {
    status?: number;
}): PlainResponse;
/** Return a 404 response. */
declare function notFound(): PlainResponse;

interface HandlerArgs {
    req: express__default.Request;
    res: express__default.Response;
}
/**
 * Defines a handler function with proper typing.
 *
 * @param handler The handler function to be defined
 * @returns The typed handler function
 */
declare function defineHandler(handler: (args: HandlerArgs) => Promise<ExpressResponse | PlainResponse | JSX.Element | JSONSerializable>): Handler;
type JSONSerializable = string | number | boolean | null | JSONSerializable[] | {
    [key: string]: JSONSerializable;
};
type Handler = (args: HandlerArgs) => Promise<ExpressResponse | PlainResponse | JSX.Element | JSONSerializable>;
type ExpressResponse = () => void;

type FileRoute = {
    filePath: string;
    routePath: string;
};
type LoadedFileRoute = {
    filePath: string;
    GET?: Handler;
    POST?: Handler;
};
type FileRouterOpts = {
    dir: string;
    ignorePatterns?: string[];
    fileRoutes?: FileRoute[];
    loadedFileRoutes?: LoadedFileRoute[];
};
/** Return an express router that serves files as routes form the given directory. */
declare function fileRouter$1(opts: FileRouterOpts): Promise<express__default.RequestHandler>;

/** Print all express routes to the console. Useful for debugging. */
declare function printRoutes<T extends Record<string, unknown>>(config: ExpandedPlainwebConfig<T, MailConfig | undefined>): Promise<void>;

type Config = ExpandedPlainwebConfig<Record<string, unknown>>;
declare function flyHeaders(): express__default.RequestHandler;
declare function migrations(config: Pick<Config, "database" | "paths" | "nodeEnv">): express__default.RequestHandler;
declare function forceWWW(config: Pick<Config, "nodeEnv">): express__default.RequestHandler;
declare function logging(config: Pick<Config, "nodeEnv">): express__default.RequestHandler;
declare function error(config: Pick<Config, "nodeEnv">): express__default.ErrorRequestHandler;
declare function staticFiles({ nodeEnv, dir, }: {
    nodeEnv: "development" | "production" | "test";
    dir: string;
}): express__default.RequestHandler;
declare function rateLimit({ nodeEnv, rateLimit, }: {
    nodeEnv: "development" | "production" | "test";
    rateLimit?: {
        windowMs?: number;
        limit?: number;
        message?: string;
    };
}): express__default.RequestHandler;
declare function redirect({ redirects, }: {
    redirects: Record<string, string>;
}): express__default.RequestHandler;
declare function json(): express__default.RequestHandler;
declare function urlencoded(): express__default.RequestHandler;
declare function database<T extends Record<string, unknown>>({ database, }: {
    database: BetterSQLite3Database<T>;
}): express__default.RequestHandler;
declare function fileRouter({ dir, }: {
    dir: string;
}): Promise<express__default.RequestHandler>;
declare function id(): express__default.RequestHandler;
/** A collection of built-in express middleware. */
declare const middleware: {
    flyHeaders: typeof flyHeaders;
    forceWWW: typeof forceWWW;
    logging: typeof logging;
    error: typeof error;
    redirect: typeof redirect;
    staticFiles: typeof staticFiles;
    rateLimit: typeof rateLimit;
    json: typeof json;
    urlencoded: typeof urlencoded;
    database: typeof database;
    fileRouter: typeof fileRouter;
    migrations: typeof migrations;
    id: typeof id;
};
/**
 * Define a middleware stack.
 * The order of middleware matters, the first middleware in the stack will be executed first.
 * */
declare function defineMiddleware(stack: (opts: MiddlewareStackArgs) => Promise<void> | void): (opts: MiddlewareStackArgs) => Promise<void> | void;

/** Run a handler with a mock request, returning the response. Useful for testing. */
declare function testHandler(handler: Handler, req: Request, { database, }: {
    database?: BetterSQLite3Database<Record<string, unknown>>;
}): Promise<node_mocks_http.MockResponse<express.Response<any, Record<string, any>>>>;

/** Return the express app instance given a plainweb config. */
declare function getApp<T extends Record<string, unknown>>(config: ExpandedPlainwebConfig<T, MailConfig | undefined>): Promise<express__default.Express>;

/**
 * This outbox traps sent emails during testing (NODE_ENV=test).
 * Access the outbox during testing to assert on sent emails.
 */
declare const outbox: JSONTransport.SentMessageInfo[];
/** Send an email using the configured mailer. */
declare function sendMail(config: MailerConfig, mailOptions: Mail.Options): Promise<void>;

/** Return the drizzle database instance given a plainweb config. */
declare function getDatabase<T extends Record<string, unknown>>(config: Pick<ExpandedPlainwebConfig<T>, "database" | "nodeEnv">): BetterSQLite3Database<T>;

/**
 * Run a a function in a database transaction.
 * The transaction is automatically rolled back, even if the function doesn't throw an error.
 * ONLY USE during testing, to keep test cases isolated from each other.
 * */
declare function isolate<Schema extends Record<string, unknown>>(db: BetterSQLite3Database<Schema>, fn: (db: BetterSQLite3Database<Schema>) => Promise<void>): Promise<void>;

/**
 * Run pending daatabasae migrations given a plainweb config.
 * Migrations change the database schema.
 */
declare function migrate(config: Pick<ExpandedPlainwebConfig<Record<string, unknown>>, "database" | "paths" | "nodeEnv">): Promise<void>;

type ErrorHandler = ({ err }: {
    err: unknown;
}) => Promise<void>;
type Task<T> = {
    name: string;
    start(opts?: {
        error?: ErrorHandler;
    }): NodeJS.Timeout;
    perform(data: T): Promise<void>;
};

/**
 * Enqueue a task to perform it later.
 * Return a promise when the task is enqueued, note that the task may not be performed immediately.
 * */
declare function perform<T>(task: Task<T>, data?: T extends void ? undefined : T): Promise<void>;

/**
 * Return a worker instance given a plainweb config.
 * When a worker is started, it starts pulling tasks from the database and performing them.
 * */
declare function getWorker(config: Pick<ExpandedPlainwebConfig<Record<string, unknown>>, "database" | "paths" | "nodeEnv">): {
    /** Load all tasks and work them. This must be called before any tasks can be enqueued. */
    start: () => Promise<Record<string, NodeJS.Timeout>>;
};

interface DefineJobOpts<T> {
    name: string;
    process: ({ data }: {
        data: T;
    }) => Promise<void>;
    success?: ({ data }: {
        data: T;
    }) => Promise<void>;
    failure?: ({ data, err }: {
        data: T;
        err: unknown;
    }) => Promise<void>;
    pollIntervall?: number;
    batchSize?: number;
    maxRetries?: number;
    retryIntervall?: number;
}
type Job<T> = {
    name: string;
    process: ({ data }: {
        data: T;
    }) => Promise<void>;
    success?: ({ data }: {
        data: T;
    }) => Promise<void>;
    failure?: ({ data, err }: {
        data: T;
        err: unknown;
    }) => Promise<void>;
};
declare function defineJob<T>(opts: DefineJobOpts<T>): Job<T>;

interface DefineTaskOpts<T> {
    name: string;
    process: ({ data }: {
        data: T;
    }) => Promise<void>;
    success?: ({ data }: {
        data: T;
    }) => Promise<void>;
    failure?: ({ data, err }: {
        data: T;
        err: unknown;
    }) => Promise<void>;
    pollIntervall?: number;
    batchSize?: number;
    maxRetries?: number;
    retryIntervall?: number;
}

declare const schema: {
    tasks: drizzle_orm_sqlite_core.SQLiteTableWithColumns<{
        name: "tasks";
        schema: undefined;
        columns: {
            id: drizzle_orm_sqlite_core.SQLiteColumn<{
                name: "id";
                tableName: "tasks";
                dataType: "string";
                columnType: "SQLiteText";
                data: string;
                driverParam: string;
                notNull: true;
                hasDefault: false;
                isPrimaryKey: true;
                isAutoincrement: false;
                hasRuntimeDefault: false;
                enumValues: [string, ...string[]];
                baseColumn: never;
                generated: undefined;
            }, object>;
            name: drizzle_orm_sqlite_core.SQLiteColumn<{
                name: "name";
                tableName: "tasks";
                dataType: "string";
                columnType: "SQLiteText";
                data: string;
                driverParam: string;
                notNull: true;
                hasDefault: false;
                isPrimaryKey: false;
                isAutoincrement: false;
                hasRuntimeDefault: false;
                enumValues: [string, ...string[]];
                baseColumn: never;
                generated: undefined;
            }, object>;
            data: drizzle_orm_sqlite_core.SQLiteColumn<{
                name: "data";
                tableName: "tasks";
                dataType: "json";
                columnType: "SQLiteTextJson";
                data: unknown;
                driverParam: string;
                notNull: false;
                hasDefault: false;
                isPrimaryKey: false;
                isAutoincrement: false;
                hasRuntimeDefault: false;
                enumValues: undefined;
                baseColumn: never;
                generated: undefined;
            }, object>;
            created: drizzle_orm_sqlite_core.SQLiteColumn<{
                name: "created";
                tableName: "tasks";
                dataType: "number";
                columnType: "SQLiteInteger";
                data: number;
                driverParam: number;
                notNull: true;
                hasDefault: false;
                isPrimaryKey: false;
                isAutoincrement: false;
                hasRuntimeDefault: false;
                enumValues: undefined;
                baseColumn: never;
                generated: undefined;
            }, object>;
            failedLast: drizzle_orm_sqlite_core.SQLiteColumn<{
                name: "failed_last";
                tableName: "tasks";
                dataType: "number";
                columnType: "SQLiteInteger";
                data: number;
                driverParam: number;
                notNull: false;
                hasDefault: false;
                isPrimaryKey: false;
                isAutoincrement: false;
                hasRuntimeDefault: false;
                enumValues: undefined;
                baseColumn: never;
                generated: undefined;
            }, object>;
            failedNr: drizzle_orm_sqlite_core.SQLiteColumn<{
                name: "failed_nr";
                tableName: "tasks";
                dataType: "number";
                columnType: "SQLiteInteger";
                data: number;
                driverParam: number;
                notNull: false;
                hasDefault: false;
                isPrimaryKey: false;
                isAutoincrement: false;
                hasRuntimeDefault: false;
                enumValues: undefined;
                baseColumn: never;
                generated: undefined;
            }, object>;
            failedError: drizzle_orm_sqlite_core.SQLiteColumn<{
                name: "failed_error";
                tableName: "tasks";
                dataType: "string";
                columnType: "SQLiteText";
                data: string;
                driverParam: string;
                notNull: false;
                hasDefault: false;
                isPrimaryKey: false;
                isAutoincrement: false;
                hasRuntimeDefault: false;
                enumValues: [string, ...string[]];
                baseColumn: never;
                generated: undefined;
            }, object>;
        };
        dialect: "sqlite";
    }>;
};
type Database = BaseSQLiteDatabase<"sync", BetterSqlite3Database.RunResult, typeof schema, ExtractTablesWithRelations<typeof schema>>;
/**
 * Define a task that gets persisted to the database.
 * A database tasks survives server restarts.
 * */
declare function defineDatabaseTask<T>(database: Database, opts: DefineTaskOpts<T>): Task<T>;

/** Define an inmemory task, DOES NOT survive server restarts. */
declare function defineInmemoryTask<T>(opts: DefineTaskOpts<T>): Task<T>;

declare function adminRouter<T extends Record<string, unknown>>(opts: {
    database: BetterSQLite3Database<T>;
    path: string;
    verbose?: number;
}): Promise<express__default.Router>;

declare const unstable_admin: typeof adminRouter;

export { type Handler, type Job, defineConfig, defineHandler, defineInmemoryTask, defineJob, defineMiddleware, defineDatabaseTask as defineTask, fileRouter$1 as fileRouter, getApp, getDatabase, getLogger, getWorker, html, isolate, json$1 as json, log, middleware, migrate, notFound, outbox, perform, printRoutes, randomId, redirect$1 as redirect, sendMail, stream, testHandler, unstable_admin };
